# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest
from magnum.objects import fields
from magnum.tests.unit.objects import utils


class TestDriver:
    @pytest.fixture(autouse=True)
    def setup(self, mocker, context, mock_pykube, mock_validate_nodegroup):
        self.cluster = utils.get_test_cluster(context, labels={})
        self.cluster.save = mocker.MagicMock()

        self.node_group = utils.get_test_nodegroup(context)
        self.node_group.save = mocker.MagicMock()

        self.mock_cluster_resource = mocker.MagicMock()
        self.mock_cluster_objects = mocker.patch(
            "magnum_cluster_api.objects.Cluster.objects"
        )
        self.mock_cluster_objects.return_value.get.return_value = (
            self.mock_cluster_resource
        )

        self.mock_wait_capi_cluster_reconciliation_start = mocker.patch(
            "magnum_cluster_api.driver.BaseDriver.wait_capi_cluster_reconciliation_start"
        )
        self.mock_get_machine_deployment = mocker.patch(
            "magnum_cluster_api.resources.get_machine_deployment"
        )

    def _assert_node_group_crud_calls(self):
        self.mock_cluster_objects.return_value.get.assert_called_once_with(
            name=self.cluster.stack_id
        )
        self.mock_cluster_resource.api.patch.assert_called_once()

    def _assert_node_group_status(self, expected_status):
        assert self.node_group.status == expected_status
        self.node_group.save.assert_called_once()

        assert self.cluster.status == fields.ClusterStatus.UPDATE_IN_PROGRESS
        self.cluster.save.assert_called_once()

    def test_create_nodegroup(self, mocker, context, ubuntu_driver):
        self.mock_cluster_resource.obj = {
            "spec": {
                "topology": {
                    "workers": {
                        "machineDeployments": [],
                    }
                }
            },
        }

        mock_mutate_machine_deployment = mocker.patch(
            "magnum_cluster_api.resources.mutate_machine_deployment"
        )

        # NOTE(okozachenko1203): We need to mock `json.dumps` because MagicMock is not JSON serializable
        mocker.patch("json.dumps", return_value="mocked_json_string")

        ubuntu_driver.create_nodegroup(context, self.cluster, self.node_group)

        mock_mutate_machine_deployment.assert_called_once_with(
            context, self.cluster, self.node_group
        )

        assert self.mock_cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ] == [mock_mutate_machine_deployment.return_value]

        self._assert_node_group_crud_calls()
        self.mock_get_machine_deployment.assert_called_once()
        self._assert_node_group_status(fields.ClusterStatus.CREATE_IN_PROGRESS)

    def test_update_nodegroup(self, mocker, context, ubuntu_driver):
        self.mock_cluster_resource.obj = {
            "spec": {
                "topology": {
                    "workers": {
                        "machineDeployments": [
                            {
                                "name": self.node_group.name,
                            }
                        ],
                    }
                }
            },
        }

        mock_mutate_machine_deployment = mocker.patch(
            "magnum_cluster_api.resources.mutate_machine_deployment"
        )

        # NOTE(okozachenko1203): We need to mock `json.dumps` because MagicMock is not JSON serializable
        mocker.patch("json.dumps", return_value="mocked_json_string")

        ubuntu_driver.update_nodegroup(context, self.cluster, self.node_group)

        mock_mutate_machine_deployment.assert_called_once_with(
            context,
            self.cluster,
            self.node_group,
            {
                "name": self.node_group.name,
            },
        )

        assert self.mock_cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ] == [mock_mutate_machine_deployment.return_value]

        self._assert_node_group_crud_calls()
        self.mock_wait_capi_cluster_reconciliation_start.assert_called_once()
        self._assert_node_group_status(fields.ClusterStatus.UPDATE_IN_PROGRESS)

    def test_update_nodegroup_with_multiple_node_groups(
        self, mocker, context, ubuntu_driver
    ):
        mock_machine_deployment = mocker.MagicMock()

        self.mock_cluster_resource.obj = {
            "spec": {
                "topology": {
                    "workers": {
                        "machineDeployments": [
                            mock_machine_deployment,
                            {
                                "name": self.node_group.name,
                            },
                        ],
                    }
                }
            },
        }

        mock_mutate_machine_deployment = mocker.patch(
            "magnum_cluster_api.resources.mutate_machine_deployment"
        )

        # NOTE(okozachenko1203): We need to mock `json.dumps` because MagicMock is not JSON serializable
        mocker.patch("json.dumps", return_value="mocked_json_string")

        ubuntu_driver.update_nodegroup(context, self.cluster, self.node_group)

        assert not mock_machine_deployment.called

        mock_mutate_machine_deployment.assert_called_once_with(
            context,
            self.cluster,
            self.node_group,
            {
                "name": self.node_group.name,
            },
        )

        assert self.mock_cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ] == [
            mock_machine_deployment,
            mock_mutate_machine_deployment.return_value,
        ]

        self._assert_node_group_crud_calls()
        self.mock_wait_capi_cluster_reconciliation_start.assert_called_once()
        self._assert_node_group_status(fields.ClusterStatus.UPDATE_IN_PROGRESS)

    def test_delete_nodegroup_with_multiple_node_groups(
        self, mocker, context, ubuntu_driver
    ):
        mock_machine_deployment = mocker.MagicMock()

        self.mock_cluster_resource.obj = {
            "spec": {
                "topology": {
                    "workers": {
                        "machineDeployments": [
                            mock_machine_deployment,
                            {
                                "name": self.node_group.name,
                            },
                        ],
                    }
                }
            },
        }

        self.mock_get_machine_deployment.return_value = None

        # NOTE(okozachenko1203): We need to mock `json.dumps` because MagicMock is not JSON serializable
        mocker.patch("json.dumps", return_value="mocked_json_string")

        ubuntu_driver.delete_nodegroup(context, self.cluster, self.node_group)

        assert not mock_machine_deployment.called

        assert self.mock_cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ] == [
            mock_machine_deployment,
        ]

        self._assert_node_group_crud_calls()
        self.mock_get_machine_deployment.assert_called_once()
        self._assert_node_group_status(fields.ClusterStatus.DELETE_IN_PROGRESS)



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import pytest
from magnum import objects as magnum_objects  # type: ignore
from magnum.objects import fields  # type: ignore

from magnum_cluster_api import clients, objects, resources


class TestNodeGroupDriver:
    @pytest.fixture(autouse=True)
    def setup(self, cluster):
        self.api = clients.get_pykube_api()
        self.cluster = cluster

    def _assert_machine_deployment_config_matches_node_group(self, md, node_group):
        assert md is not None
        # TODO: more?

    def _assert_machine_deployments_for_node_groups(
        self, *node_groups: magnum_objects.NodeGroup
    ):
        mds = objects.MachineDeployment.objects(
            self.api, namespace="magnum-system"
        ).filter(
            selector={
                "cluster.x-k8s.io/cluster-name": self.cluster.stack_id,
            },
        )

        assert len(mds) == len(
            node_groups
        ), "Expected %d MachineDeployments, got %d" % (
            len(node_groups),
            len(mds),
        )

        # NOTE(mnaser): We need to loop over all the node groups and make sure
        #               that the machine deployments are created for them.
        for ng in node_groups:
            md = resources.get_machine_deployment(self.api, self.cluster, ng)
            self._assert_machine_deployment_config_matches_node_group(md, ng)

        # NOTE(mnaser): We also need to make sure there are no extra machine
        #               deployments created.
        for md in mds:
            assert md.labels["topology.cluster.x-k8s.io/deployment-name"] in [
                ng.name for ng in node_groups
            ]

    def test_default_node_group(self):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def _create_node_group(self, context, driver, node_group_obj, node_group_name):
        new_node_group = copy.deepcopy(node_group_obj)
        new_node_group.name = node_group_name

        driver.create_nodegroup(context, self.cluster, new_node_group)

        assert new_node_group.status == fields.ClusterStatus.CREATE_IN_PROGRESS
        assert new_node_group.save.called_once()

        assert self.cluster.status == fields.ClusterStatus.UPDATE_IN_PROGRESS
        assert self.cluster.save.called_once()

        return new_node_group

    def test_create_node_group(
        self, mock_validate_nodegroup, context, ubuntu_driver, node_group_obj
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-cpu"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            new_node_group,
        )

    def test_create_and_delete_node_group(
        self, mock_validate_nodegroup, context, ubuntu_driver, node_group_obj
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-cpu"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def test_create_two_node_groups(
        self, mock_validate_nodegroup, context, ubuntu_driver, node_group_obj
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-cpu"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-memory"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )

    def test_create_and_delete_two_node_groups_deleting_newest_first(
        self, mock_validate_nodegroup, context, ubuntu_driver, node_group_obj
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-cpu"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-memory"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, second_new_node_group)
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, first_new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)

    def test_create_and_delete_two_node_groups_deleting_oldest_first(
        self, mock_validate_nodegroup, context, ubuntu_driver, node_group_obj
    ):
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)
        first_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-cpu"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
        )
        second_new_node_group = self._create_node_group(
            context, ubuntu_driver, node_group_obj, "high-memory"
        )
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            first_new_node_group,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, first_new_node_group)
        self._assert_machine_deployments_for_node_groups(
            *self.cluster.nodegroups,
            second_new_node_group,
        )
        ubuntu_driver.delete_nodegroup(context, self.cluster, second_new_node_group)
        self._assert_machine_deployments_for_node_groups(*self.cluster.nodegroups)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import io

import pykube
import yaml
from oslo_serialization import base64

from magnum_cluster_api import exceptions


class NamespacedAPIObject(pykube.objects.NamespacedAPIObject):
    @property
    def events(self):
        return pykube.Event.objects(self.api, namespace=self.namespace).filter(
            field_selector={
                "involvedObject.name": self.name,
                "involvedObject.apiVersion": self.version,
                "involvedObject.kind": self.kind,
            },
        )


class EndpointSlice(NamespacedAPIObject):
    version = "discovery.k8s.io/v1"
    endpoint = "endpointslices"
    kind = "EndpointSlice"


class ClusterResourceSet(NamespacedAPIObject):
    version = "addons.cluster.x-k8s.io/v1beta1"
    endpoint = "clusterresourcesets"
    kind = "ClusterResourceSet"


class OpenStackMachineTemplate(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackmachinetemplates"
    kind = "OpenStackMachineTemplate"


class KubeadmConfigTemplate(NamespacedAPIObject):
    version = "bootstrap.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmconfigtemplates"
    kind = "KubeadmConfigTemplate"


class KubeadmControlPlane(NamespacedAPIObject):
    version = "controlplane.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmcontrolplanes"
    kind = "KubeadmControlPlane"


class KubeadmControlPlaneTemplate(NamespacedAPIObject):
    version = "controlplane.cluster.x-k8s.io/v1beta1"
    endpoint = "kubeadmcontrolplanetemplates"
    kind = "KubeadmControlPlaneTemplate"


class MachineDeployment(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "machinedeployments"
    kind = "MachineDeployment"


class Machine(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "machines"
    kind = "Machine"


class OpenStackClusterTemplate(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackclustertemplates"
    kind = "OpenStackClusterTemplate"


class OpenStackCluster(NamespacedAPIObject):
    version = "infrastructure.cluster.x-k8s.io/v1alpha7"
    endpoint = "openstackclusters"
    kind = "OpenStackCluster"

    @property
    def identity_ref(self):
        return self.obj["spec"]["identityRef"]

    @property
    def identity_ref_secret(self) -> pykube.Secret:
        return pykube.Secret.objects(self.api, namespace=self.namespace).get(
            name=self.identity_ref["name"]
        )

    @property
    def clouds_yaml(self):
        return yaml.safe_load(
            base64.decode_as_text(self.identity_ref_secret.obj["data"]["clouds.yaml"])
        )

    @property
    def cloud_config(self):
        return self.clouds_yaml["clouds"]["default"]

    @property
    def floating_network_id(self):
        try:
            return self.obj["status"]["externalNetwork"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterExternalNetworkNotReady()

    @property
    def network_id(self):
        try:
            return self.obj["status"]["network"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterNetworkNotReady()

    @property
    def subnet_id(self):
        try:
            return self.obj["status"]["network"]["subnet"]["id"]
        except KeyError:
            raise exceptions.OpenStackClusterSubnetNotReady()

    @property
    def cloud_controller_manager_config(self):
        config = configparser.ConfigParser()
        config["Global"] = {
            "auth-url": self.cloud_config["auth"]["auth_url"],
            "region": self.cloud_config["region_name"],
            "application-credential-id": self.cloud_config["auth"][
                "application_credential_id"
            ],
            "application-credential-secret": self.cloud_config["auth"][
                "application_credential_secret"
            ],
            "tls-insecure": "false" if self.cloud_config["verify"] else "true",
        }
        config["LoadBalancer"] = {
            "floating-network-id": self.floating_network_id,
            "network-id": self.network_id,
            "subnet-id": self.subnet_id,
        }

        fd = io.StringIO()
        config.write(fd)

        return fd.getvalue()


class ClusterClass(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "clusterclasses"
    kind = "ClusterClass"


class Cluster(NamespacedAPIObject):
    version = "cluster.x-k8s.io/v1beta1"
    endpoint = "clusters"
    kind = "Cluster"

    @property
    def openstack_cluster(self):
        filtered_clusters = (
            OpenStackCluster.objects(self.api, namespace=self.namespace)
            .filter(selector={"cluster.x-k8s.io/cluster-name": self.name})
            .all()
        )

        if len(filtered_clusters) == 0:
            raise exceptions.OpenStackClusterNotCreated()

        return list(filtered_clusters)[0]


class StorageClass(pykube.objects.APIObject):
    version = "storage.k8s.io/v1"
    endpoint = "storageclasses"
    kind = "StorageClass"



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from datetime import datetime, timedelta
from unittest import mock

import openstack
import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from magnum.objects import fields
from oslo_utils import uuidutils  # type: ignore

openstack.enable_logging(debug=True)


@pytest.fixture(scope="session")
def image():
    return {
        "id": uuidutils.generate_uuid(),
        "os_distro": "ubuntu",
    }


@pytest.fixture(scope="session")
def kube_tag():
    return os.getenv("KUBE_TAG", "v1.25.3")


@pytest.fixture()
def cluster_template(mocker, kube_tag, image):
    cluster_template = mocker.MagicMock()

    cluster_template.image_id = image["id"]

    cluster_template.master_lb_enabled = True
    cluster_template.external_network_id = uuidutils.generate_uuid()
    cluster_template.dns_nameserver = "8.8.8.8"

    cluster_template.master_flavor_id = uuidutils.generate_uuid()
    cluster_template.flavor_id = uuidutils.generate_uuid()

    cluster_template.labels = {
        "kube_tag": kube_tag,
    }

    return cluster_template


@pytest.fixture()
def node_group_obj(mocker, cluster_template):
    node_group = mocker.MagicMock()
    node_group.name = "default-worker"
    node_group.role = "worker"
    node_group.node_count = 1
    node_group.flavor_id = cluster_template.flavor_id
    node_group.image_id = cluster_template.image_id
    node_group.labels = {}
    node_group.status = fields.ClusterStatus.CREATE_IN_PROGRESS

    return node_group


@pytest.fixture()
def cluster_obj(mocker, cluster_template, node_group_obj):
    cluster = mocker.MagicMock()
    cluster.project_id = uuidutils.generate_uuid()
    cluster.uuid = uuidutils.generate_uuid()
    cluster.fixed_network = None
    cluster.fixed_subnet = None
    cluster.keypair = None
    cluster.labels = {}

    cluster.cluster_template = cluster_template
    cluster.master_lb_enabled = cluster_template.master_lb_enabled
    cluster.master_flavor_id = cluster_template.master_flavor_id
    cluster.flavor_id = cluster_template.flavor_id

    cluster.master_count = 1

    cluster.default_ng_master.image_id = cluster_template.image_id
    cluster.nodegroups = [node_group_obj]

    return cluster


@pytest.fixture(scope="session")
def mock_osc(session_mocker, image):
    mock_clients = session_mocker.patch(
        "magnum_cluster_api.clients.OpenStackClients"
    ).return_value

    # NOTE(mnaser): Since there are reference in the Magnum code to the
    #               OpenStackClients, we need to make sure we mock it
    #               in the Magnum code as well.
    session_mocker.patch(
        "magnum.common.clients.OpenStackClients",
        return_value=mock_clients,
    )

    # Keystone
    mock_keystone_client = mock_clients.keystone.return_value.client
    mock_keystone_client.application_credentials.create.return_value = (
        openstack.identity.v3.application_credential.ApplicationCredential(
            id="fake_id", secret="fake_secret"
        )
    )

    # Glance
    mock_glance_client = mock_clients.glance.return_value
    mock_glance_client.images.get.return_value = image

    # Cinder
    mock_cinder_client = mock_clients.cinder.return_value
    mock_cinder_client.volume_types.default.return_value.name = "__DEFAULT__"
    mock_clients.cinder_region_name.return_value = "RegionOne"

    # Others
    mock_clients.url_for.return_value = "http://fake_url"

    return mock_clients


@pytest.fixture(scope="session")
def mock_certificate() -> mock.MagicMock:
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    cert: x509.Certificate = (
        x509.CertificateBuilder()
        .subject_name(x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, "test")]))
        .issuer_name(x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, "test")]))
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now())
        .not_valid_after(datetime.now() + timedelta(days=365))
        .sign(key, hashes.SHA256(), default_backend())
    )

    fake_certificate = mock.MagicMock()
    fake_certificate.get_certificate.return_value = cert.public_bytes(
        serialization.Encoding.PEM
    )
    fake_certificate.get_private_key_passphrase.return_value = "fake_passphrase"
    fake_certificate.get_private_key.return_value = key.private_bytes(
        serialization.Encoding.PEM,
        serialization.PrivateFormat.PKCS8,
        serialization.BestAvailableEncryption(
            fake_certificate.get_private_key_passphrase.return_value.encode("utf-8")
        ),
    )

    return fake_certificate


@pytest.fixture(scope="session")
def mock_get_cluster_ca_certificate(session_mocker, mock_certificate) -> mock.MagicMock:
    mock_get_cluster_ca_certificate = session_mocker.patch(
        "magnum.conductor.handlers.common.cert_manager.get_cluster_ca_certificate"
    )
    mock_get_cluster_ca_certificate.return_value = mock_certificate

    return mock_get_cluster_ca_certificate


@pytest.fixture(scope="session")
def mock_get_cluster_magnum_cert(session_mocker, mock_certificate) -> mock.MagicMock:
    mock_get_cluster_ca_certificate = session_mocker.patch(
        "magnum.conductor.handlers.common.cert_manager.get_cluster_magnum_cert"
    )
    mock_get_cluster_ca_certificate.return_value = mock_certificate

    return mock_get_cluster_ca_certificate


@pytest.fixture(scope="session")
def mock_certificates(mock_get_cluster_ca_certificate, mock_get_cluster_magnum_cert):
    pass


@pytest.fixture
def cluster(
    context,
    cluster_obj,
    ubuntu_driver,
    mock_validate_cluster,
    mock_osc,
    mock_certificates,
):
    try:
        ubuntu_driver.create_cluster(context, cluster_obj, 60)
        ubuntu_driver.wait_capi_cluster_reconciliation_start(context, cluster_obj, 1)

        yield cluster_obj
    finally:
        ubuntu_driver.delete_cluster(context, cluster_obj)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import getpass
import json
import os
import subprocess
import tarfile
import tempfile
import textwrap
import zlib
from os.path import expanduser
from pathlib import Path

import click
import requests

QEMU_PACKAGES = [
    "qemu-kvm",
    "qemu-utils",
]


def validate_version(_, __, value):
    if not value.startswith("v"):
        raise click.BadParameter("Version should start with 'v'")
    return value


@click.command()
@click.pass_context
@click.option(
    "--operating-system",
    show_default=True,
    default="ubuntu-2204",
    type=click.Choice(
        ["ubuntu-2004", "ubuntu-2204", "flatcar", "rockylinux-8", "rockylinux-9"]
    ),
    help="Operating system to build image for",
    prompt="Operating system to build image for",
)
@click.option(
    "--version",
    show_default=True,
    default="v1.27.8",
    callback=validate_version,
    help="Kubernetes version",
)
@click.option(
    "--image-builder-version",
    show_default=True,
    default="v0.1.25",
    help="Image builder tag (or commit) to use for building image",
)
def main(ctx: click.Context, operating_system, version, image_builder_version):
    ib_path = f"/tmp/image-builder-{image_builder_version}"
    output_path = f"{ib_path}/images/capi/output"

    # Scan the entire output directory recursively and stop if there is any file
    # at all
    for root, dirs, files in os.walk(output_path):
        if files:
            message = (
                "There are files in the output directory which will cause the build to fail. "
                "Please remove them before continuing.\n"
            )
            for file in files:
                message += f"- {root}/{file}\n"

            ctx.fail(message)

    click.echo("- Update apt")
    subprocess.run(
        ["sudo", "/usr/bin/apt-get", "update", "-y"],
        check=True,
    )

    click.echo("- Install QEMU packages")
    subprocess.run(
        ["sudo", "/usr/bin/apt", "install", "--no-install-recommends", "-y"]
        + QEMU_PACKAGES,
        check=True,
    )

    click.echo("- Add current user to KVM group")
    subprocess.run(
        ["sudo", "/usr/sbin/usermod", "-a", "-G", "kvm", getpass.getuser()], check=True
    )

    click.echo("- Update permissions for the KVM device")
    subprocess.run(["sudo", "/bin/chown", "root:kvm", "/dev/kvm"], check=True)

    # Setup our stream decompressor
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)

    click.echo("- Download image-builder")
    path = f"{ib_path}.tar"
    with requests.get(
        f"https://github.com/kubernetes-sigs/image-builder/tarball/{image_builder_version}",
        stream=True,
    ) as r:
        r.raise_for_status()
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                rv = dec.decompress(chunk)
                if rv:
                    f.write(rv)

    click.echo("- Extract image-builder")
    with tarfile.open(path) as tar:
        members = []
        for member in tar.getmembers():
            p = Path(member.path)
            member.path = p.relative_to(*p.parts[:1])
            members.append(member)

        tar.extractall(ib_path, members=members)

    click.echo("- Create customization file")
    kubernetes_series = ".".join(version.split(".")[0:2])
    customization = {
        "kubernetes_deb_version": f"{version.replace('v', '')}-1.1",
        "kubernetes_rpm_version": f"{version.replace('v', '')}",
        "kubernetes_semver": f"{version}",
        "kubernetes_series": f"{kubernetes_series}",
        # https://github.com/flatcar/Flatcar/issues/823
        "ansible_user_vars": "oem_id=openstack",
    }

    # NOTE(mnaser): We use the latest tested daily ISO for Ubuntu 22.04 in order
    #               to avoid a lengthy upgrade process.
    if operating_system == "ubuntu-2204":
        iso = "jammy-live-server-amd64.iso"

        customization["iso_url"] = (
            f"http://cdimage.ubuntu.com/ubuntu-server/jammy/daily-live/current/{iso}"
        )

        # Get the SHA256 sum for the ISO
        r = requests.get(
            "http://cdimage.ubuntu.com/ubuntu-server/jammy/daily-live/current/SHA256SUMS"
        )
        r.raise_for_status()
        for line in r.text.splitlines():
            if iso in line:
                customization["iso_checksum"] = line.split()[0]
                break

        # Assert that we have the checksum
        assert "iso_checksum" in customization

    if operating_system == "rockylinux-8":
        iso = "Rocky-x86_64-minimal.iso"

        customization["iso_url"] = (
            f"https://download.rockylinux.org/pub/rocky/8/isos/x86_64/{iso}"
        )

        # Get the SHA256 sum for the ISO
        r = requests.get(
            "https://download.rockylinux.org/pub/rocky/8/isos/x86_64/Rocky-x86_64-minimal.iso.CHECKSUM"
        )
        r.raise_for_status()
        for line in r.text.splitlines():
            if iso in line and "SHA256" in line:
                customization["iso_checksum"] = line.split()[3]
                break

        # Assert that we have the checksum
        assert "iso_checksum" in customization

    if operating_system == "rockylinux-9":
        iso = "Rocky-x86_64-minimal.iso"

        customization["iso_url"] = (
            f"https://download.rockylinux.org/pub/rocky/9/isos/x86_64/{iso}"
        )

        # Get the SHA256 sum for the ISO
        r = requests.get(
            "https://download.rockylinux.org/pub/rocky/9/isos/x86_64/Rocky-x86_64-minimal.iso.CHECKSUM"
        )
        r.raise_for_status()
        for line in r.text.splitlines():
            if iso in line and "SHA256" in line:
                customization["iso_checksum"] = line.split()[3]
                break

        # Assert that we have the checksum
        assert "iso_checksum" in customization

    # NOTE(mnaser): Let's set number of CPUs to equal the number of CPUs on the
    #               host to speed up the build process.
    customization["cpus"] = str(os.cpu_count())

    # NOTE(mnaser): We set the memory of the VM to 50% of the total memory
    #               of the system.
    with open("/proc/meminfo", "r") as f:
        for line in f:
            if line.startswith("MemTotal:"):
                total_memory_kb = int(line.split()[1])
                break
    total_memory_mb = total_memory_kb / 1024
    customization["memory"] = str(int(total_memory_mb * 0.5))

    with tempfile.NamedTemporaryFile(suffix=".json") as fp:
        fp.write(json.dumps(customization).encode("utf-8"))
        fp.flush()

        click.echo("- Build image")
        subprocess.run(
            [
                "/usr/bin/newgrp",
                "kvm",
            ],
            input=textwrap.dedent(
                f"""\
                /usr/bin/make \
                -C \
                {ib_path}/images/capi \
                build-qemu-{operating_system}
            """
            ).encode("utf-8"),
            env={
                **os.environ,
                **{
                    "PATH": f"{expanduser('~/.local/bin')}:{os.environ['PATH']}",
                    "PACKER_VAR_FILES": fp.name,
                },
            },
        )

    # Try and detect the target image path since it can be different depending
    # on the operating system, so we scan the output directory for the file
    # that matches the pattern.
    target = None
    for root, dirs, files in os.walk(output_path):
        if len(files) > 1:
            ctx.fail(f"Unexpected number of files in {root}")
        if files:
            target = f"{root}/{files[0]}"

    if target is None:
        ctx.fail("Unable to detect target image")

    # Copy from the target to the current working directory
    os.rename(target, f"{operating_system}-kube-{version}.qcow2")



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from __future__ import annotations

import keystoneauth1
from magnum import objects as magnum_objects
from magnum.conductor import scale_manager
from magnum.drivers.common import driver
from magnum.objects import fields
from tenacity import Retrying, retry, stop_after_delay, wait_fixed

from magnum_cluster_api import (
    clients,
    exceptions,
    monitor,
    objects,
    resources,
    sync,
    utils,
)


def cluster_lock_wrapper(func):
    def wrapper(*args, **kwargs):
        cluster = args[2]  # Assuming cluster is the second argument
        with sync.ClusterLock(cluster.uuid):
            return func(*args, **kwargs)

    return wrapper


class BaseDriver(driver.Driver):
    def __init__(self):
        self.k8s_api = clients.get_pykube_api()

    def create_cluster(
        self, context, cluster: magnum_objects.Cluster, cluster_create_timeout: int
    ):
        # NOTE(mnaser): We want to set the `stack_id` as early as possible to
        #               make sure we can use it in the cluster creation.
        cluster.stack_id = utils.generate_cluster_api_name(self.k8s_api)
        cluster.save()

        utils.validate_cluster(context, cluster)
        resources.Namespace(self.k8s_api).apply()

        return self._create_cluster(context, cluster)

    @cluster_lock_wrapper
    def _create_cluster(self, context, cluster: magnum_objects.Cluster):
        osc = clients.get_openstack_api(context)

        credential = osc.keystone().client.application_credentials.create(
            user=cluster.user_id,
            name=cluster.uuid,
            description=f"Magnum cluster ({cluster.uuid})",
        )

        resources.CloudConfigSecret(
            context,
            self.k8s_api,
            cluster,
            osc.cinder_region_name(),
            credential,
        ).apply()

        resources.ApiCertificateAuthoritySecret(context, self.k8s_api, cluster).apply()
        resources.EtcdCertificateAuthoritySecret(context, self.k8s_api, cluster).apply()
        resources.FrontProxyCertificateAuthoritySecret(
            context, self.k8s_api, cluster
        ).apply()
        resources.ServiceAccountCertificateAuthoritySecret(
            context, self.k8s_api, cluster
        ).apply()

        resources.apply_cluster_from_magnum_cluster(
            context, self.k8s_api, cluster, skip_auto_scaling_release=True
        )

    def _get_cluster_status_reason(self, capi_cluster):
        capi_cluster_status_reason = ""
        capi_ops_cluster_status_reason = ""

        # Get the latest event message of the CAPI Cluster
        capi_cluster_events = capi_cluster.events
        if capi_cluster_events:
            capi_cluster_status_reason += utils.format_event_message(
                list(capi_cluster_events)[-1]
            )

        # Get the latest event message of the CAPI OpenstackCluster
        capi_ops_cluster_events = []
        capi_ops_cluster = capi_cluster.openstack_cluster
        if capi_ops_cluster:
            capi_ops_cluster_events = capi_ops_cluster.events
        if capi_ops_cluster_events:
            capi_ops_cluster_status_reason += utils.format_event_message(
                list(capi_ops_cluster_events)[-1]
            )

        return "CAPI Cluster status: %s. CAPI OpenstackCluster status reason: %s" % (
            capi_cluster_status_reason,
            capi_ops_cluster_status_reason,
        )

    @cluster_lock_wrapper
    def update_cluster_status(
        self, context, cluster: magnum_objects.Cluster, use_admin_ctx: bool = False
    ):
        # NOTE(mnaser): We may be called with a stale cluster object, so we
        #               need to refresh it to make sure we have the latest data.
        cluster.refresh()

        node_groups = [
            self.update_nodegroup_status(context, cluster, node_group)
            for node_group in cluster.nodegroups
        ]
        # TODO: watch for topology change instead
        osc = clients.get_openstack_api(context)

        capi_cluster = resources.Cluster(context, self.k8s_api, cluster).get_or_none()

        if cluster.status in (
            fields.ClusterStatus.CREATE_IN_PROGRESS,
            fields.ClusterStatus.UPDATE_IN_PROGRESS,
        ):
            # NOTE(mnaser): It's possible we run a cluster status update before
            #               the cluster is created. In that case, we don't want
            #               to update the cluster status.
            if capi_cluster is None:
                return

            capi_cluster.reload()
            status_map = {
                c["type"]: c["status"] for c in capi_cluster.obj["status"]["conditions"]
            }

            for condition in ("ControlPlaneReady", "InfrastructureReady", "Ready"):
                if status_map.get(condition) != "True":
                    cluster.status_reason = self._get_cluster_status_reason(
                        capi_cluster
                    )
                    cluster.save()
                    return

            api_endpoint = capi_cluster.obj["spec"]["controlPlaneEndpoint"]
            cluster.api_address = (
                f"https://{api_endpoint['host']}:{api_endpoint['port']}"
            )
            cluster.coe_version = capi_cluster.obj["spec"]["topology"]["version"]

            # NOTE(oleks): To avoid autoscaler crashes, we deploy it after the
            #              cluster api endpoint is reachable.
            if (
                cluster.status == fields.ClusterStatus.CREATE_IN_PROGRESS
                and utils.get_auto_scaling_enabled(cluster)
            ):
                resources.ClusterAutoscalerHelmRelease(self.k8s_api, cluster).apply()

            for ng in node_groups:
                if not ng.status.endswith("_COMPLETE"):
                    return
                if ng.status == fields.ClusterStatus.DELETE_COMPLETE:
                    ng.destroy()

            if cluster.status == fields.ClusterStatus.CREATE_IN_PROGRESS:
                cluster.status_reason = None
                cluster.status = fields.ClusterStatus.CREATE_COMPLETE
            if cluster.status == fields.ClusterStatus.UPDATE_IN_PROGRESS:
                cluster.status_reason = None
                cluster.status = fields.ClusterStatus.UPDATE_COMPLETE

            cluster.save()

        if cluster.status == fields.ClusterStatus.DELETE_IN_PROGRESS:
            if capi_cluster and capi_cluster.exists():
                cluster.status_reason = self._get_cluster_status_reason(capi_cluster)
                cluster.save()
                return

            # NOTE(mnaser): We delete the application credentials at this stage
            #               to make sure CAPI doesn't lose access to OpenStack.
            try:
                osc.keystone().client.application_credentials.find(
                    name=cluster.uuid,
                    user=cluster.user_id,
                ).delete()
            except keystoneauth1.exceptions.http.NotFound:
                pass

            resources.CloudConfigSecret(context, self.k8s_api, cluster).delete()
            resources.ApiCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.EtcdCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.FrontProxyCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()
            resources.ServiceAccountCertificateAuthoritySecret(
                context, self.k8s_api, cluster
            ).delete()

            cluster.status_reason = None
            cluster.status = fields.ClusterStatus.DELETE_COMPLETE
            cluster.save()

    @cluster_lock_wrapper
    def update_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        scale_manager=None,
        rollback=False,
    ):
        raise NotImplementedError()

    @cluster_lock_wrapper
    def resize_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        resize_manager: scale_manager.ScaleManager,
        node_count: int,
        nodes_to_remove: list[str],
        nodegroup: magnum_objects.NodeGroup = None,
    ):
        utils.validate_cluster(context, cluster)

        if nodes_to_remove:
            machines = objects.Machine.objects(self.k8s_api).filter(
                namespace="magnum-system",
                selector={
                    "cluster.x-k8s.io/cluster-name": cluster.stack_id,
                    "topology.cluster.x-k8s.io/deployment-name": nodegroup.name,
                },
            )

            for machine in machines:
                instance_uuid = machine.obj["spec"]["providerID"].split("/")[-1]
                if instance_uuid in nodes_to_remove:
                    machine.obj["metadata"].setdefault("annotations", {})
                    machine.obj["metadata"]["annotations"][
                        "cluster.x-k8s.io/delete-machine"
                    ] = "yes"
                    utils.kube_apply_patch(machine)

        self._update_nodegroup(context, cluster, nodegroup)

    @cluster_lock_wrapper
    def upgrade_cluster(
        self,
        context,
        cluster: magnum_objects.Cluster,
        cluster_template: magnum_objects.ClusterTemplate,
        max_batch_size,
        nodegroup: magnum_objects.NodeGroup,
        scale_manager=None,
        rollback=False,
    ):
        """
        Upgrade a cluster to a new version of Kubernetes.
        """
        # TODO: nodegroup?

        # Get current generation
        current_generation = resources.Cluster(
            context, self.k8s_api, cluster
        ).get_observed_generation()
        resources.apply_cluster_from_magnum_cluster(
            context, self.k8s_api, cluster, cluster_template=cluster_template
        )
        # Wait till the generation has been increased
        self.wait_capi_cluster_reconciliation_start(
            context, cluster, current_generation
        )

        # NOTE(mnaser): We need to save the cluster status here to make sure
        #               it happens inside the lock.
        cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        cluster.save()

    @retry(
        stop=stop_after_delay(10),
        wait=wait_fixed(1),
    )
    def wait_capi_cluster_reconciliation_start(
        self, context, cluster: magnum_objects.Cluster, old_generation: int
    ):
        """Wait until the cluster's new generation is observed by capi-controller

        This means the cluster reconciliation has been started and the conditions has been updated.
        """
        current_generation = resources.Cluster(
            context, self.k8s_api, cluster
        ).get_observed_generation()
        if old_generation != current_generation:
            return
        raise exceptions.ClusterAPIReconcileTimeout()

    @cluster_lock_wrapper
    def delete_cluster(self, context, cluster: magnum_objects.Cluster):
        if cluster.stack_id is None:
            return
        # NOTE(mnaser): This should be removed when this is fixed:
        #
        #               https://github.com/kubernetes-sigs/cluster-api-provider-openstack/issues/842
        #               https://github.com/kubernetes-sigs/cluster-api-provider-openstack/pull/990
        utils.delete_loadbalancers(context, cluster)

        resources.ClusterResourceSet(self.k8s_api, cluster).delete()
        resources.ClusterResourcesConfigMap(context, self.k8s_api, cluster).delete()
        resources.Cluster(context, self.k8s_api, cluster).delete()
        resources.ClusterAutoscalerHelmRelease(self.k8s_api, cluster).delete()

    @cluster_lock_wrapper
    def create_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        utils.validate_nodegroup(nodegroup, context)

        cluster_resource: objects.Cluster = objects.Cluster.objects(
            self.k8s_api, namespace="magnum-system"
        ).get(name=cluster.stack_id)

        cluster_resource.obj["spec"]["topology"]["workers"][
            "machineDeployments"
        ].append(resources.mutate_machine_deployment(context, cluster, nodegroup))

        utils.kube_apply_patch(cluster_resource)

        for attempt in Retrying(stop=stop_after_delay(10), wait=wait_fixed(1)):
            with attempt:
                md = resources.get_machine_deployment(self.k8s_api, cluster, nodegroup)
                if md is None:
                    raise exceptions.ClusterAPIReconcileTimeout()

        nodegroup.status = fields.ClusterStatus.CREATE_IN_PROGRESS
        nodegroup.save()

        cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        cluster.save()

    def update_nodegroup_status(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        action = nodegroup.status.split("_")[0]

        if nodegroup.role == "master":
            kcp = resources.get_kubeadm_control_plane(self.k8s_api, cluster)
            if kcp is None:
                return nodegroup

            generation = kcp.obj.get("status", {}).get("observedGeneration", 1)
            if generation > 1:
                action = "UPDATE"

            ready = kcp.obj["status"].get("ready", False)
            failure_message = kcp.obj["status"].get("failureMessage")

            updated_replicas = kcp.obj["status"].get("updatedReplicas")
            replicas = kcp.obj["status"].get("replicas")

            if updated_replicas != replicas:
                nodegroup.status = f"{action}_IN_PROGRESS"
            elif updated_replicas == replicas and ready:
                nodegroup.status = f"{action}_COMPLETE"
            nodegroup.status_reason = failure_message
        else:
            md = resources.get_machine_deployment(self.k8s_api, cluster, nodegroup)
            if md is None:
                if action == "DELETE":
                    nodegroup.status = f"{action}_COMPLETE"
                    nodegroup.save()
                    return nodegroup
                return nodegroup

            phase = md.obj["status"]["phase"]

            if phase in ("ScalingUp", "ScalingDown"):
                nodegroup.status = f"{action}_IN_PROGRESS"
            elif phase == "Running":
                nodegroup.status = f"{action}_COMPLETE"
            elif phase in ("Failed", "Unknown"):
                nodegroup.status = f"{action}_FAILED"

        nodegroup.save()

        return nodegroup

    @cluster_lock_wrapper
    def update_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        self._update_nodegroup(context, cluster, nodegroup)

    def _update_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        utils.validate_nodegroup(nodegroup, context)

        cluster_resource: objects.Cluster = objects.Cluster.objects(
            self.k8s_api, namespace="magnum-system"
        ).get(name=cluster.stack_id)

        machine_deployment_index = None
        for i, machine_deployment in enumerate(
            cluster_resource.obj["spec"]["topology"]["workers"]["machineDeployments"]
        ):
            if machine_deployment["name"] == nodegroup.name:
                machine_deployment_index = i
                break

        if machine_deployment_index is not None:
            machine_deployment = cluster_resource.obj["spec"]["topology"]["workers"][
                "machineDeployments"
            ][machine_deployment_index]

            cluster_resource.obj["spec"]["topology"]["workers"]["machineDeployments"][
                machine_deployment_index
            ] = resources.mutate_machine_deployment(
                context,
                cluster,
                nodegroup,
                machine_deployment,
            )

        current_generation = resources.Cluster(
            context, self.k8s_api, cluster
        ).get_observed_generation()
        utils.kube_apply_patch(cluster_resource)
        self.wait_capi_cluster_reconciliation_start(
            context, cluster, current_generation
        )

        nodegroup.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        nodegroup.save()

        cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        cluster.save()

    @cluster_lock_wrapper
    def delete_nodegroup(
        self,
        context,
        cluster: magnum_objects.Cluster,
        nodegroup: magnum_objects.NodeGroup,
    ):
        cluster_resource: objects.Cluster = objects.Cluster.objects(
            self.k8s_api, namespace="magnum-system"
        ).get(name=cluster.stack_id)

        machine_deployment_index = None
        for i, machine_deployment in enumerate(
            cluster_resource.obj["spec"]["topology"]["workers"]["machineDeployments"]
        ):
            if machine_deployment["name"] == nodegroup.name:
                machine_deployment_index = i
                break

        if machine_deployment_index is not None:
            del cluster_resource.obj["spec"]["topology"]["workers"][
                "machineDeployments"
            ][machine_deployment_index]

        utils.kube_apply_patch(cluster_resource)

        for attempt in Retrying(stop=stop_after_delay(10), wait=wait_fixed(1)):
            with attempt:
                md = resources.get_machine_deployment(self.k8s_api, cluster, nodegroup)
                if md is not None:
                    raise exceptions.ClusterAPIReconcileTimeout()

        nodegroup.status = fields.ClusterStatus.DELETE_IN_PROGRESS
        nodegroup.save()

        cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
        cluster.save()

    @cluster_lock_wrapper
    def get_monitor(self, context, cluster: magnum_objects.Cluster):
        return monitor.Monitor(context, cluster)

    # def rotate_ca_certificate(self, context, cluster):
    #     raise exception.NotSupported(
    #         "'rotate_ca_certificate' is not supported by this driver.")

    def create_federation(self, context, federation):
        raise NotImplementedError()

    def update_federation(self, context, federation):
        raise NotImplementedError()

    def delete_federation(self, context, federation):
        raise NotImplementedError()


class UbuntuDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "ubuntu", "coe": "kubernetes"},
        ]


class UbuntuFocalDriver(UbuntuDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "ubuntu-focal", "coe": "kubernetes"},
        ]


class FlatcarDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "flatcar", "coe": "kubernetes"},
        ]


class RockyLinuxDriver(BaseDriver):
    @property
    def provides(self):
        return [
            {"server_type": "vm", "os": "rockylinux", "coe": "kubernetes"},
        ]
